import sys

from .dataprep import dataprep
from .diffmod import diffmod
from .postprocessing import postprocessing

def parse_options(argv):

    """Parses options from the command line """

    from argparse import ArgumentParser
    from xpore import __version__

    parser = ArgumentParser(prog='xpore')
    subparsers = parser.add_subparsers(help='Running modes', metavar='{dataprep, diffmod, postprocessing}')
    parser.add_argument('-v', '--version', action='version', version='%(prog)s {version}'.format(version=__version__))

    ### RUN MODE "DATAPREP"
    parser_dataprep = subparsers.add_parser('dataprep', help='run mode for preprocessing nanopolish eventalign.txt before differential modification analysis')
    optional_dataprep = parser_dataprep._action_groups.pop()
    required_dataprep = parser_dataprep.add_argument_group('required arguments')
    # Required arguments
    required_dataprep.add_argument('--eventalign', dest='eventalign', help='eventalign filepath, the output from nanopolish.',required=True)
    ##required.add_argument('--summary', dest='summary', help='eventalign summary filepath, the output from nanopolish.',required=True)
    required_dataprep.add_argument('--out_dir', dest='out_dir', help='output directory.',required=True)
    optional_dataprep.add_argument('--gtf_or_gff', dest='gtf_or_gff', help='GTF or GFF file path.',type=str)
    optional_dataprep.add_argument('--transcript_fasta', dest='transcript_fasta', help='transcript FASTA path.',type=str)
    # Optional arguments
    optional_dataprep.add_argument('--skip_eventalign_indexing', dest='skip_eventalign_indexing', help='skip indexing the eventalign nanopolish output.',default=False,action='store_true')
    # parser.add_argument('--features', dest='features', help='Signal features to extract.',type=list,default=['norm_mean'])
    optional_dataprep.add_argument('--genome', dest='genome', help='to run on Genomic coordinates. Without this argument, the program will run on transcriptomic coordinates',default=False,action='store_true') 
    optional_dataprep.add_argument('--n_processes', dest='n_processes', help='number of processes to run.',type=int, default=1)
    optional_dataprep.add_argument('--chunk_size', dest='chunk_size', help='number of lines from nanopolish eventalign.txt for processing.',type=int, default=1000000)
    optional_dataprep.add_argument('--readcount_min', dest='readcount_min', help='minimum read counts per gene.',type=int, default=1)
    optional_dataprep.add_argument('--readcount_max', dest='readcount_max', help='maximum read counts per gene.',type=int, default=1000)
    optional_dataprep.add_argument('--resume', dest='resume', help='with this argument, the program will resume from the previous run.',default=False,action='store_true') #todo
    parser_dataprep._action_groups.append(optional_dataprep)
    parser_dataprep.set_defaults(func=dataprep)

    ### RUN MODE "DIFFMOD"
    parser_diffmod = subparsers.add_parser('diffmod', help='run mode for performing differential modification analysis')
    optional_diffmod = parser_diffmod._action_groups.pop()
    required_diffmod = parser_diffmod.add_argument_group('required arguments')
    # Required arguments
    required_diffmod.add_argument('--config', dest='config', help='YAML configuraion filepath.',required=True)
    # Optional arguments
    optional_diffmod.add_argument('--n_processes', dest='n_processes', help='number of processes to run.',type=int,default=1)
    optional_diffmod.add_argument('--save_models', dest='save_models', help='with this argument, the program will save the model parameters for each id.',default=False,action='store_true') # todo
    optional_diffmod.add_argument('--resume', dest='resume', help='with this argument, the program will resume from the previous run.',default=False,action='store_true') 
    optional_diffmod.add_argument('--ids', dest='ids', help='gene or transcript ids to model.',default=[],nargs='*')
    parser_diffmod._action_groups.append(optional_diffmod)
    parser_diffmod.set_defaults(func=diffmod)

    ### RUN MODE "POSTPROCESSING"
    parser_postprocessing = subparsers.add_parser('postprocessing', help='run mode for post-processing diffmod.table, the result table from differential modification analysis.')
    required_postprocessing = parser_postprocessing.add_argument_group('required arguments')
    # Required arguments
    required_postprocessing.add_argument('--diffmod_dir', dest='diffmod_dir', help='diffmod directory path, the output from xpore-diffmod.',required=True)
    parser_postprocessing.set_defaults(func=postprocessing)

    return parser.parse_args(argv[1:])

def main(argv=sys.argv):

    ### get command line options
    options = parse_options(argv)
    options.func(options)

if __name__ == "__main__":
    main(sys.argv)



import yaml
import os
from collections import defaultdict

from ..utils import misc

def get_condition_run_name(condition_name,run_name):
    return '-'.join([condition_name,run_name])

class Configurator(object):
    def __init__(self, config_filepath):
        self.filepath = os.path.abspath(config_filepath)
        self.filename = self.filepath.split('/')[-1]
        self.yaml = yaml.safe_load(open(self.filepath, 'r'))
        
    def get_paths(self):
        paths = {}
        
        if 'prior' in self.yaml:
            paths['model_kmer'] = os.path.abspath(self.yaml['prior'])
        else:
            paths['model_kmer'] = os.path.join(os.path.dirname(__file__),'model_kmer.csv')

        paths['out_dir'] = os.path.join(os.path.abspath(self.yaml['out']))
        paths.update(misc.makedirs(paths['out_dir'],sub_dirs=['models']))
        paths['model_filepath'] = os.path.join(paths['out_dir'], 'models', '%s.model')        
        return paths
        
    def get_data_info(self):
        data = defaultdict(dict)
        for condition_name, run_names in self.yaml['data'].items():
            for run_name, dirpath in run_names.items():
                data[condition_name][get_condition_run_name(condition_name,run_name)] = dirpath
        return data
    
    def get_criteria(self):
        criteria = {}
        if 'criteria' in self.yaml.keys():
            criteria = self.yaml['criteria']
        else:
            criteria['readcount_min'] = 15
            criteria['readcount_max'] = 1000

        return criteria
        
    def get_method(self):
        if 'method' in self.yaml.keys():
            method = self.yaml['method']
        else:
            method = {}

        method.setdefault('name', 'gmm')
        method.setdefault('max_iters', 500)
        method.setdefault('stopping_criteria', 0.00001)
        method.setdefault('compute_elbo', True)
        method.setdefault('verbose', False)
        method.setdefault('update', ['z','y','w','mu_tau'])
        method.setdefault('pooling', False)
        method.setdefault('prefiltering',False)
        return method
    
    def get_priors(self):
        prior_params = defaultdict(dict)
        if 'priors' not in self.yaml.keys():
            # mu_tau
            prior_params['mu_tau']['location'] = ['model_kmer_mean','model_kmer_mean']
            prior_params['mu_tau']['lambda'] = [1,1]
            prior_params['mu_tau']['alpha'] = [0.5,0.5]
            prior_params['mu_tau']['beta'] = ['model_kmer_tau','model_kmer_tau']
            prior_params['mu_tau']['beta_scale'] = [0.5,0.5]

            # w
            prior_params['w']['concentration'] = [0.001,0.001]        
        else:
            pass #todo

        return prior_params
        



import numpy as np
from numpy import newaxis
import scipy.special
import scipy.stats


class GMM(object):  
    """
    1D multi-sample 2-Gaussian mixture model.
    """
    def __init__(self, method=None, data={'x': None, 'y': None, 'condition_names': None, 'run_names': None}, inits={'info': None, 'nodes': {'x': None, 'y': None, 'w': None, 'mu_tau': None, 'z': None}}, priors={'mu_tau': None, 'w': None},kmer_signal=None):
        """
        Parameters
        ----------
        method: dict
            Config file
        data: dict
            Sth.
        inits: dict
            Sth.
        priors:
            Sth.
        """
        
        self.nodes = dict()
        self.aux_params = dict()
        self.K = 2  # modified and unmodified
        self.method = method
        self.kmer_signal = kmer_signal
        
        self.__init_info(inits['info'])

        data_node_x = None
        if (data['x'] is not None) and (data['y'] is not None) and (data['r'] is not None):
            
            self.info['n_reads'] = len(data['y'])

            # inits
            location_0 = priors['mu_tau']['location'][0]
            if location_0 < np.percentile(data['y'], q=50):
                location_1 = np.percentile(data['y'], q=75)
            else:
                location_1 = np.percentile(data['y'], q=25)
            locations = np.array([location_0, location_1])
            # print(locations)

            # kmeans = KMeans(n_clusters=self.K).fit(data['y'][:,newaxis])
            # locations = kmeans.cluster_centers_.flatten()

            inits['nodes']['mu_tau'] = {'location': locations, 'lambda': priors['mu_tau']['lambda'], 'alpha': priors['mu_tau']['alpha'], 'beta': priors['mu_tau']['beta']}
            
            if self.method['pooling']:
                inits['nodes']['x'] = {'group_names': data['condition_names']}
                data_node_x = data['x']
            else:
                inits['nodes']['x'] = {'group_names': data['run_names']}
                data_node_x = data['r']

            # additional nodes for postprocessing analysis
            self.nodes['y_condition_names'] = Constant(data=data['y_condition_names'])
            self.nodes['y_run_names'] = Constant(data=data['y_run_names'])

        # Define the graph
        self.nodes['x'] = Constant(data=data_node_x, inits=inits['nodes']['x'])   # NG
        self.n_groups = len(self.nodes['x'].params['group_names'])
        self.nodes['w'] = Dirichlet(dim=(self.n_groups, self.K), inits=inits['nodes']['w'], priors=priors['w'])
        self.nodes['mu_tau'] = UnivariateNormalGamma(dim=(self.K), inits=inits['nodes']['mu_tau'], priors=priors['mu_tau'])
        self.nodes['z'] = Bernoulli(dim=(self.info['n_reads'], self.K), parents={'w': self.nodes['w'], 'x': self.nodes['x']}, inits=inits['nodes']['z'])
        self.nodes['y'] = UnivariateNormalMixture(parents={'z': self.nodes['z'], 'mu_tau': self.nodes['mu_tau'], 'x': self.nodes['x']}, data=data['y'], inits=inits['nodes']['y'])

    def __init_info(self, inits):
        if inits is None:
            self.info = {'n_reads': 0, 'log_elbos': [], 'converged': False, 'n_iterations': -1, 'convergence_ratio': -1}
        else:
            self.info = inits

    def __compute_log_elbo(self):
        """
        Compute log ELBO.
        """
        log_elbo = self.nodes['y']._log_likelihood()
        log_elbo += self.nodes['z']._log_prob_prior()
        log_elbo -= self.nodes['z']._log_prob_posterior()
        log_elbo += self.nodes['w']._log_prob_prior()
        log_elbo -= self.nodes['w']._log_prob_posterior()
        log_elbo += self.nodes['mu_tau']._log_prob_prior(var='normal')
        log_elbo += self.nodes['mu_tau']._log_prob_prior(var='gamma')
        log_elbo -= self.nodes['mu_tau']._log_prob_posterior(var='normal')
        log_elbo -= self.nodes['mu_tau']._log_prob_posterior(var='gamma')

        return log_elbo

    def fit(self):
        """
        Fit.
        """
        error = False
        converged = False
        # self.__init_params()
        log_elbo_old = self.__compute_log_elbo()  # -np.inf
        self.nodes['y']._update()

        for iteration in range(1, self.method['max_iters']):
            self.nodes['z']._update(children={'y': self.nodes['y']})  # every time z is updated, y needs to be done too. Update z so that z is not randomly drawn.
            self.nodes['y']._update()
            self.nodes['w']._update(children=self.nodes['y'].params)
            self.nodes['mu_tau']._update(children={**self.nodes['y'].params, **{'x': self.nodes['x']}})

            log_elbo_new = self.__compute_log_elbo()
            error = np.round(log_elbo_new - log_elbo_old, 8) < 0
            if error:
                print('ERROR: log ELBO is decreasing ...')
                print(iteration, log_elbo_old, log_elbo_new, error)
                error = True
                break
            else:
                self.info['log_elbos'] += [log_elbo_old]
                diff = (log_elbo_new - log_elbo_old)/np.abs(log_elbo_old)
                log_elbo_old = log_elbo_new
                # print('Iteration %d:' %(iteration),'Convergence ratio =',diff)
                if (diff < self.method['stopping_criteria']):
                    converged = True
                    break
        self.info['n_iterations'] = iteration
        self.info['converged'] = converged
        self.info['convergence_ratio'] = diff
        return self

################################
##### Model for each node ######
################################


class Constant(object):
    """
    Constant node.
    """
    def __init__(self, data=None, inits=None):
        self.params = inits
        self.data = data


class UnivariateNormalMixture(object):
    def __init__(self, parents=None, data=None, inits=None):

        self.parents = parents
        self.data = data

        self.params = dict.fromkeys(['N', 'mean', 'variance'])
        self.__initialise_params(inits)

    def __initialise_params(self, inits):
        if inits is not None:
            self.params = inits

    def _log_likelihood(self):
        n_reads = len(self.data)
        mu_tau_nomal_means = np.repeat(self.parents['mu_tau'].expected(var='normal')[newaxis, :], repeats=n_reads, axis=-2)  # N,K
        mu_tau_normal_variances = np.repeat(self.parents['mu_tau'].variance(var='normal')[newaxis, :], repeats=n_reads, axis=-2)  # N,K
        mu_tau_gamma_means = np.repeat(self.parents['mu_tau'].expected(var='gamma')[newaxis, :], repeats=n_reads, axis=-2)  # N,K
        mu_tau_log_gamma = np.repeat(self.parents['mu_tau'].expected(var='gamma', inner_func='log')[newaxis, :], repeats=n_reads, axis=-2)  # N,K
        y = self.data[..., newaxis]  # N,1

        log_likelihood = self.parents['z'].expected() * 0.5 * (mu_tau_log_gamma - np.log(2*np.pi) - mu_tau_gamma_means * (y**2 - 2*y * mu_tau_nomal_means + mu_tau_nomal_means**2 + mu_tau_normal_variances))  # N,K
        log_likelihood *= (np.sum(self.parents['x'].data, axis=-1)[..., newaxis] > 0)  # N,K
        return np.sum(np.sum(log_likelihood, axis=-1), axis=-1)  # sum across components and reads respectively => 1

    def expected(self, inner_func=None):
        return self.params['mean']

    def variance(self, inner_func=None):
        return self.params['variance']

    def N(self, inner_func=None):
        return self.params['N']

    def _update(self):
        n_reads = len(self.data)
        self.params['N'] = np.sum(self.parents['x'].data[..., newaxis] * self.parents['z'].expected()[:, newaxis, :], axis=-3)  # sum across reads => G,K
        N = np.sum(self.params['N'], axis=-2)  # sum across groups => K
        N[N < 1e-30] = 0.  # Hack: Avoid overflow in later calculation steps.

        N_inverse = np.divide(1., N, out=np.zeros_like(N), where=N != 0.)
        y = self.data[..., newaxis]  # N,1
        self.params['mean'] = N_inverse * np.sum(self.parents['z'].expected()*y, axis=-2)  # sum across reads => K

        residuals = (np.sum(self.parents['x'].data, axis=-1) > 0)[..., newaxis]*(y - np.repeat(self.params['mean'][newaxis, :], repeats=n_reads, axis=-2))  # N,K

        self.params['variance'] = N_inverse * np.sum(self.parents['z'].expected()*(residuals**2), axis=-2)  # sum across reads => K


class Bernoulli(object):
    def __init__(self, dim, parents=None, data=None, inits=None):
        self.params = dict()
        self.params['prob'] = np.full(dim, np.nan)
        self.params['ln_prob'] = np.full(dim, np.nan)
        self.__initialise_params(inits)

        self.parents = parents
        self.data = data

    def __initialise_params(self, inits=None):
        if inits is None:
            # self.params['prob'][:] = 0.5
            # self.params['ln_prob'][:] = np.log(0.5)

            self.params['prob'] = np.random.random(self.params['prob'].shape)
            self.params['prob'] /= np.sum(self.params['prob'], axis=1)[:, newaxis]
            self.params['ln_prob'] = np.log(self.params['prob'])
        else:
            self.params = inits

    def _log_prob_prior(self):
        res = self.expected() * np.sum(self.parents['x'].data[..., newaxis] * self.parents['w'].expected(inner_func='log')[newaxis, :, :], axis=-2)  # sum across groups => N,K
        return np.sum(np.sum(res, axis=-1), axis=-1)  # 1

    def _log_prob_posterior(self):
        res = self.expected() * np.sum(self.parents['x'].data[:, newaxis, :] * self.params['ln_prob'][..., newaxis], axis=-1)  # N,K
        return np.sum(np.sum(res, axis=-1), axis=-1)  # 1

    def expected(self, inner_func=None):
        return self.params['prob']

    # def variance(self):
    #     return self.params['prob']*(1-self.params['prob']) #scipy.stats.bernoulli.var(self.params['concentration']) #

    def _update(self, children):
        # print('#'*5,'Updating z','#'*5)
        n_reads = len(children['y'].data)
        mu_tau_nomal_means = np.repeat(children['y'].parents['mu_tau'].expected(var='normal')[newaxis, :], repeats=n_reads, axis=-2)  # N,K
        mu_tau_normal_variances = np.repeat(children['y'].parents['mu_tau'].variance(var='normal')[newaxis, :], repeats=n_reads, axis=-2)  # N,K
        mu_tau_gamma_means = np.repeat(children['y'].parents['mu_tau'].expected(var='gamma')[newaxis, :], repeats=n_reads, axis=-2)  # N,K
        mu_tau_expected_log_gamma = np.repeat(children['y'].parents['mu_tau'].expected(var='gamma', inner_func='log')[newaxis, :], repeats=n_reads, axis=-2)  # N,K

        y = children['y'].data[..., newaxis]  # N,1
        ln_rho = 0.5*(mu_tau_expected_log_gamma - np.log(2*np.pi) - mu_tau_gamma_means * (y**2 - 2*y * mu_tau_nomal_means + mu_tau_nomal_means**2 + mu_tau_normal_variances))

        ln_rho += np.sum(self.parents['x'].data[..., newaxis]*self.parents['w'].expected(inner_func='log')[newaxis, :, :], axis=-2)  # sum across groups => N,K

        ln_rho_diff = ln_rho[:, 1]-ln_rho[:, 0]
        prob = 1./(1.+np.exp(np.clip(ln_rho_diff, -500, 500)))  # N,1 => N
        self.params['prob'] = np.stack([prob, 1-prob], axis=-1)
        self.params['ln_prob'] = ln_rho - scipy.special.logsumexp(ln_rho, axis=-1)[..., newaxis]  # N,K
    


class Dirichlet(object):
    def __init__(self, dim, parents=None, data=None, inits=None, priors=None):  # dim - [,n_categories]

        self.priors = dict()
        self.priors['concentration'] = np.full(dim, np.nan)
        self.__set_priors(priors)

        self.params = dict()
        self.params['concentration'] = np.full(dim, np.nan)
        self.__initialise_params(inits)

        self.parents = parents
        self.data = data

    def __initialise_params(self, inits=None):
        if inits is None:
            self.params['concentration'][:] = self.priors['concentration'][:]
        else:
            self.params = inits

    def __set_priors(self, priors=None):
        if priors is None:
            self.priors['concentration'][:] = 1
        else:
            self.priors = priors

    def _log_prob_prior(self):
        res = Dirichlet.__log_C(self.priors['concentration']) + np.sum((self.priors['concentration']-1)*self.expected(inner_func='log'), axis=-1)  # sum k => G
        return np.sum(res, axis=-1)  # 1

    def _log_prob_posterior(self):
        res = Dirichlet.__log_C(self.params['concentration']) + np.sum((self.params['concentration']-1)*self.expected(inner_func='log'), axis=-1)  # sum k => G
        return np.sum(res, axis=-1)  # 1

    def expected(self, inner_func=None):
        if inner_func == 'log':
            res = scipy.special.digamma(self.params['concentration'])  # G,K
            res -= scipy.special.digamma(np.sum(self.params['concentration'], axis=-1))[..., newaxis]  # G,K / G,1 => G,K
            return res
        else:
            return self.params['concentration']/np.sum(self.params['concentration'], axis=-1)[..., newaxis]  # G,K

    def variance(self):  # To check, still wrong but unused.
        return (self.expected()*(1-self.expected())) / (np.sum(self.params['concentration'], axis=-1)[..., newaxis]+1)

    def _update(self, children):
        # print('#'*5,'Updating w','#'*5)
        self.params['concentration'] = self.priors['concentration'] + children['N']

    def __log_C(alpha):
        return scipy.special.gammaln(np.sum(alpha, axis=-1)) - np.sum(scipy.special.gammaln(alpha), axis=-1)


class UnivariateNormalGamma(object):
    def __init__(self, dim, parents=None, data=None, inits=None, priors=None):

        self.priors = dict.fromkeys(['location', 'lambda', 'alpha', 'beta'], np.full(dim, np.nan))  # alpha = shape, beta=rate
        self.__set_priors(priors)

        self.params = dict.fromkeys(['location', 'lambda', 'alpha', 'beta'], np.full(dim, np.nan))  # alpha = shape, beta=rate
        self.__initialise_params(inits)
        self.parents = parents
        self.data = data

    def __initialise_params(self, inits):
        if inits is None:
            for param in self.params:
                self.params[param] = self.priors[param][:]
            self.params['location'][:] = np.random.normal(loc=self.priors['location'][:], scale=1)

        else:
            self.params = inits

    def __set_priors(self, priors=None):
        # if priors is None:
        #     self.priors['location'][:] = 0. # Todo: set to mean
        #     self.priors['lambda'][:] = 1.
        #     self.priors['alpha'][:] = 1.
        #     self.priors['beta'][:] = 1.
        # else:
        self.priors = priors

    def _log_prob_prior(self, var='normal'):
        if var == 'normal':
            res = 0.5 * (self.expected(var='gamma', inner_func='log') + np.log(self.priors['lambda']) - np.log(2*np.pi) - self.priors['lambda']*self.expected(var='gamma') * (self.expected(var='normal')**2 + self.variance(var='normal') - 2*self.expected(var='normal')*self.priors['location'] + self.priors['location']**2))  # K

        else:  # gamma
            res = self.priors['alpha']*np.log(self.priors['beta'])-scipy.special.gammaln(self.priors['alpha'])+(self.priors['alpha']-1)*self.expected(var='gamma', inner_func='log') - self.priors['beta']*self.expected(var='gamma')  # K

        return np.sum(res, axis=-1)  # 1

    def _log_prob_posterior(self, var='normal'):
        if var == 'normal':
            res = 0.5 * (self.expected(var='gamma', inner_func='log') + np.log(self.params['lambda']) - np.log(2*np.pi) - self.params['lambda']*self.expected(var='gamma') * (self.expected(var='normal')**2 + self.variance(var='normal') - 2*self.expected(var='normal')*self.params['location'] + self.params['location']**2))  # K

        else:  # gamma
            res = self.params['alpha']*np.log(self.params['beta'])-scipy.special.gammaln(self.params['alpha'])+(self.params['alpha']-1)*self.expected(var='gamma', inner_func='log') - self.params['beta']*self.expected(var='gamma')  # K

        return np.sum(res, axis=-1)  # 1

    def expected(self, var='normal', inner_func=None):
        if inner_func == 'log':
            if var == 'gamma':  # tau
                return scipy.special.digamma(self.params['alpha']) - np.log(self.params['beta'])  # K

        else:
            if var == 'normal':
                return self.params['location']
            else:  # gamma
                return self.params['alpha'] / self.params['beta']

    def variance(self, var='normal'):
        if var == 'normal':
            return 1./(self.params['lambda']*self.expected(var='gamma'))
        # else: # gamma
        #     return self.params['alpha']/(self.params['beta']**2)

    def _update(self, children):
        # print('#'*5,'Updating mu,tau','#'*5)
        N = np.sum(children['N'], axis=-2)  # sum across groups => K
        self.params['lambda'] = self.priors['lambda'] + N  # K
        self.params['location'] = (1./self.params['lambda']) * (N*children['mean'] + self.priors['lambda']*self.priors['location'])  # K
        self.params['alpha'] = self.priors['alpha'] + 0.5*N  # K
        self.params['beta'] = self.priors['beta'] + 0.5*(N*children['variance'] + ((self.priors['lambda']*N) / (self.priors['lambda']+N))*(children['mean']-self.priors['location'])**2)  # K



import numpy as np
import pandas
import os
import multiprocessing 
import ujson
from collections import defaultdict
import csv

from . import helper
from ..diffmod.configurator import Configurator
from ..diffmod.gmm import GMM
from ..diffmod import io
from ..diffmod.statstest import StatsTest
        
def execute(idx, data_dict, data_info, method, criteria, model_kmer, prior_params, out_paths, save_models,locks):
    """
    Run the model on each posiiton across the given idx.
    """
    data = io.load_data(idx,data_dict,min_count=criteria['readcount_min'],max_count=criteria['readcount_max'],pooling=method['pooling']) 
    models = dict()
    for key,data_at_pos in data.items(): # For each position
        idx, pos, kmer = key
        kmer_signal = {'mean':model_kmer.loc[kmer,'model_mean'],'std':model_kmer.loc[kmer,'model_stdv']}
        kmer_signal['tau'] = 1./(kmer_signal['std']**2)
        y_mean = data_at_pos['y'].mean()
        y_tau = 1./(data_at_pos['y'].std()**2)

        K = 2
        if method['pooling']:
            n_groups = len(data_at_pos['condition_names'])
        else:
            n_groups = len(data_at_pos['run_names'])

        ### Set up priors.
        priors = {'mu_tau':defaultdict(list),'w':dict()}

        for k in range(K):
            priors['mu_tau']['location'] += [kmer_signal['mean']]
            priors['mu_tau']['lambda'] += [prior_params['mu_tau']['lambda'][k]]
            priors['mu_tau']['alpha'] += [kmer_signal['tau']]
            priors['mu_tau']['beta'] += [prior_params['mu_tau']['beta_scale'][k]*1./kmer_signal['tau']]
        
        for k,v in priors['mu_tau'].items():
            priors['mu_tau'][k] = np.array(v)
        
        priors['w']['concentration'] = np.ones([n_groups,K])*1. #GK
        priors['w']['concentration'][:,0] = float(prior_params['w']['concentration'][0])
        priors['w']['concentration'][:,1] = float(prior_params['w']['concentration'][1])
        ###

        ### Fit a model.
        if method['prefiltering']:
            pval = StatsTest(data_at_pos).fit(method=method['prefiltering']['method'])
            if np.isnan(pval) | (pval < method['prefiltering']['threshold']):
                prefiltering = {method['prefiltering']['method']:pval}
                models[key] = GMM(method,data_at_pos,priors=priors,kmer_signal=kmer_signal).fit(), prefiltering
        else:
            models[key] = GMM(method,data_at_pos,priors=priors,kmer_signal=kmer_signal).fit(), None

        
    if save_models & (len(models)>0): #todo: 
        print(out_paths['model_filepath'],idx)
        io.save_models_to_hdf5(models,out_paths['model_filepath'])
        
    if len(models)>0:
        # Generating the result table.
        table = io.generate_result_table(models,data_info)
        with locks['table'], open(out_paths['table'],'a') as f:
            csv.writer(f,delimiter=',').writerows(table)
        # # Logging
        # log_str = '%s: Saving the result table ... Done.' %(idx)
        # with locks['log'], open(out_paths['log'],'a') as f:
        #     f.write(log_str + '\n')
        
    # Logging
    with locks['log'], open(out_paths['log'],'a') as f:
        f.write(idx + '\n')
                        
def diffmod(args):
    
    n_processes = args.n_processes       
    config_filepath = args.config
    save_models = args.save_models
    resume = args.resume
    ids = args.ids

    config = Configurator(config_filepath) 
    paths = config.get_paths()
    data_info = config.get_data_info()
    method = config.get_method()
    criteria = config.get_criteria()
    prior_params = config.get_priors()

    print('Using the signal of unmodified RNA from',paths['model_kmer'])
    model_kmer = pandas.read_csv(paths['model_kmer']).set_index('model_kmer')
    ###

    ###
    # Get gene ids for modelling
    # todo
    
    # Create output paths and locks.
    out_paths,locks = dict(),dict()
    for out_filetype in ['model','table','log']:
        out_paths[out_filetype] = os.path.join(paths['out_dir'],'diffmod.%s' %out_filetype)
        locks[out_filetype] = multiprocessing.Lock()
        
    # Create communication queues.
    task_queue = multiprocessing.JoinableQueue(maxsize=n_processes * 2)

    # Writing the starting of the files.
    ids_done = []
    if resume and os.path.exists(out_paths['log']):
        ids_done = [line.rstrip('\n') for line in open(out_paths['log'],'r')]  
    else:
        with open(out_paths['table'],'w') as f:
            csv.writer(f,delimiter=',').writerow(io.get_result_table_header(data_info,method))
        with open(out_paths['log'],'w') as f:
            f.write(helper.decor_message('diffmod'))


    # Create and start consumers.
    consumers = [helper.Consumer(task_queue=task_queue,task_function=execute,locks=locks) for i in range(n_processes)]
    for p in consumers:
        p.start()

    ### Load tasks in to task_queue. ###
    f_index,f_data = {},{}
    for condition_name, run_names in data_info.items():
        for run_name, dirpath in run_names.items():
            # Read index files
            df_index = pandas.read_csv(os.path.join(dirpath,'data.index'),sep=',') 
            f_index[run_name] = dict(zip(df_index['idx'],zip(df_index['start'],df_index['end'])))

            # Read readcount files
            # df_readcount[run_name] = pandas.read_csv(os.path.join(info['dirpath'],'readcount.csv')).groupby('gene_id')['n_reads'].sum() # todo: data.readcount

            # Open data files
            f_data[run_name] = open(os.path.join(dirpath,'data.json'),'r') 
    
    # Load tasks into task_queue.
    # gene_ids = helper.get_gene_ids(config.filepath)
#    gene_ids = ['ENSG00000168496','ENSG00000204388','ENSG00000123989','ENSG00000170144'] #test data; todo
    # gene_ids = ['ENSG00000159111']    
    
    if len(ids) == 0:
        ids = helper.get_ids(f_index,data_info)


    print(len(ids),'ids to be testing ...')
    
    for idx in ids:
        if resume and (idx in ids_done):
            continue
        
        data_dict = dict()
        for condition_name, run_names in data_info.items():
            for run_name, dirpath in run_names.items():
                try:
                    pos_start,pos_end = f_index[run_name][idx]
                except KeyError:
                    data_dict[(condition_name,run_name)] = None
                else:
                    # print(idx,run_name,pos_start,pos_end,df_readcount[run_name].loc[idx])
                    f_data[run_name].seek(pos_start,0)
                    json_str = f_data[run_name].read(pos_end-pos_start)
                    # print(json_str[:50])
                    # json_str = '{%s}' %json_str # used for old dataprep
                    data_dict[(condition_name,run_name)] = ujson.loads(json_str) # A data dict for each gene.
                
        # tmp
        out_paths['model_filepath'] = os.path.join(paths['models'],'%s.hdf5' %idx)
        #
        # if data_dict[run_name][idx] is not None: # todo: remove this line. Fix in dataprep
        task_queue.put((idx, data_dict, data_info, method, criteria, model_kmer, prior_params, out_paths,save_models)) # Blocked if necessary until a free slot is available.

        
    # Put the stop task into task_queue.
    task_queue = helper.end_queue(task_queue,n_processes)

    # Wait for all of the tasks to finish.
    task_queue.join()

    # Close data files
    for f in f_data.values():
        f.close()   

    with open(out_paths['log'],'a+') as f:
        f.write(helper.decor_message('successfully finished'))



import numpy as np
import h5py
import os
from collections import OrderedDict, defaultdict
import itertools
import scipy.stats

from ..utils import stats
from .gmm import GMM


def get_dummies(x):
    X = []
    labels = sorted(list(set(x)))
    for label in labels:
        X += [x == label]
    # labels = [label.encode('UTF-8') for label in labels]
    return np.array(X).T, labels


def load_data(idx, data_dict, min_count, max_count, pooling=False): 
    """
    Parameters
    ----------
    data_dict: dict of ...
        Data for a gene 
    """
    
    # Create all (pos,kmer) pairs from all runs.
    position_kmer_pairs = []
    for (condition_name,run_name), d_dict in data_dict.items(): # data_dict[run_name][idx][position][kmer]
        pairs = []
        if d_dict is not None:
            for pos in d_dict[idx].keys():
                for kmer in d_dict[idx][pos].keys():
                    pairs += [(pos,kmer)]                
            position_kmer_pairs += [pairs]

    position_kmer_pairs = set(position_kmer_pairs[0]).intersection(*position_kmer_pairs)

    data = OrderedDict()
    for pos,kmer in position_kmer_pairs:  
        y, read_ids, condition_labels, run_labels = [], [], [], []
        n_reads = defaultdict(list)
        
        for (condition_name,run_name), d_dict in data_dict.items():
            if d_dict is not None:
                norm_means = d_dict[idx][pos][kmer]#['norm_means']
                n_reads_per_run = len(norm_means)
                # In case of pooling==False, if not enough reads, don't include them. 
                if (not pooling) and ((n_reads_per_run < min_count) or (n_reads_per_run > max_count)):
                    continue
                #
                n_reads[condition_name] += [n_reads_per_run]
                y += norm_means
                # read_ids += list(data_dict[run_name][idx][pos][kmer]['read_ids'][:])
                condition_labels += [condition_name]*n_reads_per_run
                run_labels += [run_name]*n_reads_per_run

        y = np.array(y)
        # read_ids = np.array(read_ids)
        condition_labels = np.array(condition_labels)
        run_labels = np.array(run_labels)
        
        # Filter those sites that don't have enough reads.
        if len(y) == 0:  # no reads at all.
            continue
        conditions_incl = []
        unique_condition_names = {condition_name for condition_name,_ in data_dict.keys()}
        if pooling: # At the modelling step all the reads from the same condition will be combined.
            for condition_name in unique_condition_names:
                if (sum(n_reads[condition_name]) >= min_count) and (sum(n_reads[condition_name]) <= max_count):
                    conditions_incl += [condition_name]
        else:
            for condition_name in unique_condition_names:
                if (np.array(n_reads[condition_name]) >= min_count).any() and (np.array(n_reads[condition_name]) <= max_count).any():
                    conditions_incl += [condition_name]
                    
        if len(conditions_incl) < 2:
            continue

        # Get dummies
        x, condition_names_dummies = get_dummies(condition_labels)
        r, run_names_dummies = get_dummies(run_labels)

        key = (idx, pos, kmer)

        data[key] = {'y': y, 'x': x, 'r': r, 'condition_names': condition_names_dummies, 'run_names': run_names_dummies, 'y_condition_names': condition_labels, 'y_run_names': run_labels} # , 'read_ids': read_ids

    return data


def save_result_table(table, out_filepath):
    out_file = h5py.File(out_filepath, 'w')
    out_file['result'] = table  # Structured np array.
    out_file.close()


def save_models_to_hdf5(models, model_filepath):  # per gene/transcript
    """
    Save model parameters.

    Parameters
    ----------
    models
        Learned models.
    model_filepath: str
        Path to save the models.
    """
    model_file = h5py.File(model_filepath, 'w')
    for model_key, model_tuple in models.items():
        idx, position, kmer = model_key
        model, prefiltering = model_tuple

        position = str(position)
        if idx not in model_file:
            model_file.create_group(idx)
        model_file[idx].create_group(position)
        model_file[idx][position].attrs['kmer'] = kmer.encode('UTF-8')
        # model_file[idx][position].create_group('info')
        # for key, value in model.info.items():
        #     model_file[idx][position]['info'][key] = value

        model_file[idx][position].create_group('nodes')  # ['x','y','z','w','mu_tau'] => store only their params
        for node_name in model.nodes:
            model_file[idx][position]['nodes'].create_group(node_name)
            if model.nodes[node_name].data is not None: # Todo: make it optional.
                value = model.nodes[node_name].data
                if node_name in ['y_run_names','y_condition_names']:
                    value = [val.encode('UTF-8') for val in value]
                model_file[idx][position]['nodes'][node_name]['data'] = value
            if model.nodes[node_name].params is None:
                continue
            for param_name, value in model.nodes[node_name].params.items():
                if param_name == 'group_names':
                    value = [val.encode('UTF-8') for val in value]
                model_file[idx][position]['nodes'][node_name][param_name] = value

    model_file.close()


def load_models(model_filepath):  # per gene/transcript #Todo: refine.
    """
    Construct a model and load model parameters.

    Parameters
    ----------
    model_filepath: str
        Path where the model is stored.

    Return
    ------
    models
        Models for each genomic position.
    """

    model_file = h5py.File(model_filepath, 'r')
    models = {}
    data = defaultdict(dict)
    for idx in model_file:
        for position in model_file[idx]:
            inits = {'info': None, 'nodes': {'x': {}, 'y': {}, 'w': {}, 'mu_tau': {}, 'z': {}}}
            kmer = model_file[idx][position].attrs['kmer']
            key = (idx, position, kmer)
            # for k in model_file[idx][position]['info']:
            #     inits['info'] = model_file[idx][position]['info'][k]
            for node_name, params in model_file[idx][position]['nodes'].items():
                for param_name, value in params.items():
                    if param_name == 'data':
                        data[key][node_name] = value[:]
                    else:
                        inits['nodes'][node_name][param_name] = value[:]
                # for param_name, value in priors.items():
                #     inits['nodes'][node_name][param_name] = value[:]

            models[key] = GMM(data,inits=inits)

    model_file.close()

    return models,data  # {(idx,position,kmer): GMM obj}

def get_ordered_condition_run_names(data_info):
    unique_condition_names = list(data_info.keys())
    unique_condition_names_sorted = sorted(unique_condition_names)
    unique_run_names_sorted = sum([list(run_names.keys()) for _, run_names in data_info.items()],[])

    return unique_condition_names_sorted, unique_run_names_sorted

def get_result_table_header(data_info,method):
    condition_names,run_names = get_ordered_condition_run_names(data_info)
    ### stats header
    stats_pairwise = []
    for cond1, cond2 in itertools.combinations(condition_names, 2):
        pair = '_vs_'.join((cond1, cond2)) # cond1 - cond2
        stats_pairwise += ['diff_mod_rate_%s' % pair, 'pval_%s' % pair, 'z_score_%s' % pair]
    stats_one_vs_all = []
    for condition_name in condition_names:
        pair = '%s_vs_all' %condition_name # condition_name - the rest
        stats_one_vs_all += ['diff_mod_rate_%s' % pair, 'pval_%s' %pair, 'z_score_%s' %pair]

    ### position header
    header = ['id', 'position', 'kmer']
    
    # header += ['p_overlap']
    # header += ['x_x1', 'y_x1', 'x_x2', 'y_x2']
    
    if method['pooling']:
        names = condition_names
    else:
        names = run_names
    
    ### stats header
    header += stats_pairwise
    if len(condition_names) > 2:
        header += stats_one_vs_all
    ###
    ### model header
    for name in names:
        header += ['mod_rate_%s' % name]
    for name in names:
        header += ['coverage_%s' % name]
    header += ['mu_unmod', 'mu_mod', 'sigma2_unmod', 'sigma2_mod', 'conf_mu_unmod', 'conf_mu_mod','mod_assignment']
    ###
    
    ### prefiltering header
    if method['prefiltering']:
        header += [method['prefiltering']['method']]
    return header

def generate_result_table(models, data_info):  # per idx (gene/transcript)
    """
    Generate a table containing learned model parameters and statistic tests.

    Parameters
    ----------
    models
        Learned models for individual genomic positions of a gene.
    group_labels
        Labels of samples.
    data_inf
        Dict

    Returns
    -------
    table
        List of tuples.
    """

    ###
    condition_names,run_names = get_ordered_condition_run_names(data_info) # information from the config file used for modelling.
    ###

    ###
    table = []
    for key, (model,prefiltering) in models.items():
        idx, position, kmer = key
        mu = model.nodes['mu_tau'].expected()  # K
        sigma2 = 1./model.nodes['mu_tau'].expected(var='gamma')  # K
        var_mu = model.nodes['mu_tau'].variance(var='normal')  # K
        # mu = model.nodes['y'].params['mean']
        # sigma2 = model.nodes['y'].params['variance']
        w = model.nodes['w'].expected()  # GK
        N = model.nodes['y'].params['N'].round()  # GK
        N0 = N[:, 0].squeeze()
        N1 = N[:, 1].squeeze()
        w0 = w[:, 0].squeeze()
        coverage = np.sum(model.nodes['y'].params['N'], axis=-1)  # GK => G # n_reads per group

        p_overlap, list_cdf_at_intersections = stats.calc_prob_overlapping(mu, sigma2)

        model_group_names = model.nodes['x'].params['group_names'] #condition_names if pooling, run_names otherwise.
        
        ### Cluster assignment ###
        conf_mu = [calculate_confidence_cluster_assignment(mu[0],model.kmer_signal),calculate_confidence_cluster_assignment(mu[1],model.kmer_signal)]
    
        cluster_idx = {}
        if conf_mu[0] > conf_mu[1]:
            cluster_idx['unmod'] = 0
            cluster_idx['mod'] = 1
        else:
            cluster_idx['unmod'] = 1
            cluster_idx['mod'] = 0

        mu_assigned = [mu[cluster_idx['unmod']],mu[cluster_idx['mod']]]
        sigma2_assigned = [sigma2[cluster_idx['unmod']],sigma2[cluster_idx['mod']]] 
        conf_mu = [conf_mu[cluster_idx['unmod']],conf_mu[cluster_idx['mod']]]
        w_mod = w[:,cluster_idx['mod']]
        mod_assignment = [['higher','lower'][(mu[0]<mu[1])^cluster_idx['mod']]]
            
        
        ### calculate stats_pairwise
        stats_pairwise = []
        for cond1, cond2 in itertools.combinations(condition_names, 2):
            if model.method['pooling']:
                cond1, cond2 = [cond1], [cond2]
            else:
                cond1, cond2 = list(data_info[cond1].keys()), list(data_info[cond2].keys())
            if any(r in model_group_names for r in cond1) and any(r in model_group_names for r in cond2):
                w_cond1 = w[np.isin(model_group_names, cond1), cluster_idx['mod']].flatten()
                w_cond2 = w[np.isin(model_group_names, cond2), cluster_idx['mod']].flatten()
                n_cond1 = coverage[np.isin(model_group_names, cond1)]
                n_cond2 = coverage[np.isin(model_group_names, cond2)]

                z_score, p_ws = stats.z_test(w_cond1, w_cond2, n_cond1, n_cond2) # two=tailed
                w_mod_mean_diff = np.mean(w_cond1)-np.mean(w_cond2)

                stats_pairwise += [w_mod_mean_diff, p_ws, z_score]
            else:
                stats_pairwise += [None, None, None]

        if len(condition_names) > 2:
            ### calculate stats_one_vs_all
            stats_one_vs_all = []
            for cond in condition_names:
                if model.method['pooling']:
                    cond = [cond]
                else:
                    cond = list(data_info[cond].keys())
                if any(r in model_group_names for r in cond):
                    w_cond1 = w[np.isin(model_group_names, cond), cluster_idx['mod']].flatten()
                    w_cond2 = w[~np.isin(model_group_names, cond), cluster_idx['mod']].flatten()
                    n_cond1 = coverage[np.isin(model_group_names, cond)]
                    n_cond2 = coverage[~np.isin(model_group_names, cond)]

                    z_score, p_ws = stats.z_test(w_cond1, w_cond2, n_cond1, n_cond2)
                    w_mod_mean_diff = np.mean(w_cond1)-np.mean(w_cond2)

                    stats_one_vs_all += [w_mod_mean_diff, p_ws, z_score]
                else:
                    stats_one_vs_all += [None, None, None]

        ###
        w_mod_ordered, coverage_ordered = [], [] # ordered by conditon_names or run_names based on headers.        
        if model.method['pooling']:
            names = condition_names
        else:
            names = run_names
        for name in names:
            if name in model_group_names:
                w_mod_ordered += list(w_mod[np.isin(model_group_names, name)])
                coverage_ordered += list(coverage[np.isin(model_group_names, name)])
            else:
                w_mod_ordered += [None]
                coverage_ordered += [None]

        ###
        ### prepare values to write
        row = [idx, position, kmer]
        row += stats_pairwise
        if len(condition_names) > 2:
            row += stats_one_vs_all

        # row += [p_overlap]
        # row += list_cdf_at_intersections
        row += list(w_mod_ordered)
        row += list(coverage_ordered)
        row += mu_assigned + sigma2_assigned + conf_mu + mod_assignment


        if prefiltering is not None:
            row += [prefiltering[model.method['prefiltering']['method']]]
        ### Filtering those positions with a nearly single distribution.
        cdf_threshold = 0.1
        x_x1, y_x1, x_x2, y_x2 = list_cdf_at_intersections
        is_not_inside = ((y_x1 < cdf_threshold) & (x_x1 < cdf_threshold)) | ((y_x2 < cdf_threshold) & (x_x2 < cdf_threshold)) | (( (1-y_x1) < cdf_threshold) & ((1-x_x1) < cdf_threshold)) | (( (1-y_x2) < cdf_threshold) & ((1-x_x2) < cdf_threshold))
        if (p_overlap <= 0.5) and (is_not_inside):
            table += [tuple(row)]

    return table


def calculate_confidence_cluster_assignment(mu,kmer_signal):
    cdf = scipy.stats.norm.cdf(kmer_signal['mean'] - abs(kmer_signal['mean']-mu), loc=kmer_signal['mean'], scale=kmer_signal['std'])
    return cdf*2



import gzip
import multiprocessing
import numpy
import os
import pandas
from functools import reduce
from collections import defaultdict


class EventalignFile:
    
    def __init__(self, fn):
        self._fn = fn
        self._open()

    def _open(self):
        fn = self._fn
        if os.path.splitext(fn)[1] == '.gz':
            self._handle = gzip.open(fn)
            self._decode_method = bytes.decode
        else:
            self._handle = open(fn)
            self._decode_method = str

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.close()

    def close(self):
        self._handle.close()

    def readline(self):
        self._handle.readline()

    def __iter__(self):
        return self

    def __next__(self):
        return self._decode_method(next(self._handle))


def decor_message(text,opt='simple'):
    text = text.upper()
    if opt == 'header':
        return text
    else:
        return '--- ' + text + ' ---\n'

def end_queue(task_queue,n_processes):
    for _ in range(n_processes):
        task_queue.put(None)
    return task_queue
        
def get_ids(f_index,data_info): #todo
    df_list = []
      
    for condition_name, run_names in data_info.items():
        list_of_set_ids = []
        for run_name in run_names:
            list_of_set_ids += [set(f_index[run_name].keys())]
        # ids = reduce(lambda x,y: x.intersection(y), list_of_set_ids)
        ids = reduce(lambda x,y: x.union(y), list_of_set_ids)
        df_list += [pandas.DataFrame({'ids':list(ids),condition_name:[1]*len(ids)})]
    df_merged = reduce(lambda  left,right: pandas.merge(left,right,on=['ids'], how='outer'), df_list).fillna(0).set_index('ids')
    return sorted(list(df_merged[df_merged.sum(axis=1) >= 2].index)) # At least two conditions.
    
## tmp: to remove
# def get_gene_ids(config_filepath): # arguments are not used.
#     import os,pandas
#     from ..diffmod.configurator import Configurator
#     # config_filepath = '/ploy_ont_workspace/github/experiments/Release_v1_0/config_manuscript/gmm_HEK293T-KO_HEK293T-WT_HEPG2-WT_K562-WT_A549-WT_MCF7-WT_reps_v01.ini'
#     config = Configurator(config_filepath) 
#     paths = config.get_paths()
#     info = config.get_info()
#     criteria = config.get_criteria() 
#     df_gt_ids = pandas.read_csv('/ploy_ont_workspace/out/Release_v1_0/statCompare/data/mapping_gt_ids.csv')    
#     gene_ids = set(df_gt_ids['g_id'].unique())
#     read_count_sum_min,read_count_sum_max = criteria['read_count_min'],criteria['read_count_max']
#     df_read_count = {}
#     for run_name in set(info['run_names']):
#         read_count_filepath = os.path.join(paths['data_dir'],run_name,'summary','read_count_per_gene.csv')
#         df_read_count[run_name] = pandas.read_csv(read_count_filepath).set_index('g_id') 

#     for run_name in set(info['run_names']):
#         df_read_count[run_name].reset_index(inplace=True)
#         cond = df_read_count[run_name]['n_reads'] >= criteria['read_count_min']
#         cond &= df_read_count[run_name]['n_reads'] <= criteria['read_count_max']
#         gene_ids = gene_ids.intersection(set(df_read_count[run_name].loc[cond,'g_id'].values))

#     gene_ids = sorted(list(gene_ids))
#     return gene_ids
##

class Consumer(multiprocessing.Process):
    """ For parallelisation """
    
    def __init__(self,task_queue,task_function,locks=None,result_queue=None):
        multiprocessing.Process.__init__(self)
        self.task_queue = task_queue
        self.locks = locks
        self.task_function = task_function
        self.result_queue = result_queue
        
    def run(self):
        proc_name = self.name
        while True:
            next_task_args = self.task_queue.get()
            if next_task_args is None:
                self.task_queue.task_done()
                break
            result = self.task_function(*next_task_args,self.locks)    
            self.task_queue.task_done()
            if self.result_queue is not None:
                self.result_queue.put(result)

def read_last_line(filepath): # https://stackoverflow.com/questions/3346430/what-is-the-most-efficient-way-to-get-first-and-last-line-of-a-text-file/3346788
    if not os.path.exists(filepath):
        return
    with open(filepath, "rb") as f:
        first = f.readline()        # Read the first line.
        if first == b'':
            return
        f.seek(-2, os.SEEK_END)     # Jump to the second last byte.
        while f.read(1) != b"\n":   # Until EOL is found...
            f.seek(-2, os.SEEK_CUR) # ...jump back the read byte plus one more.
        last = f.readline()         # Read last line.
    return last

def is_successful(filepath):
    return read_last_line(filepath) == b'--- SUCCESSFULLY FINISHED ---\n'
    


import scipy.stats
from math import fabs, sqrt, log, erf
import numpy as np
# from ..exts.statistics import NormalDist


def z_test(y1, y2, n1, n2):
    # two-tailed
    p1 = y1.mean()
    p2 = y2.mean()
    n1 = n1.mean().round()
    n2 = n2.mean().round()
    se = np.sqrt(p1*(1-p1)/n1 + p2*(1-p2)/n2)
    z = (p1 - p2) / se
    return z, scipy.stats.norm.sf(abs(z))*2


def calc_prob_overlapping(means, variances):
    sigma1, sigma2 = np.sqrt(variances)
    mu1, mu2 = means
    ### the bug causing entry had sigma1 == sigma2 and mu1 == mu2
    #print("sigma1: ",sigma1,"sigma2: ",sigma2)
    #print("mu1: ",mu1,"mu2: ",mu2)
    if sigma1 != sigma2 and mu1 != mu2:
        return NormalDist(mu=mu1, sigma=sigma1).overlap(NormalDist(mu=mu2, sigma=sigma2))
    else:
        return np.nan,[np.nan,np.nan,np.nan,np.nan]

class NormalDist(object):
    "Modified from the library statistics.py (Authors)(https://github.com/python/cpython/blob/3.8/Lib/statistics.py) which is under GPL-compatible license."
    def __init__(self, mu=0.0, sigma=1.0):
        self._mu = mu
        self._sigma = sigma
    
    def cdf(self, x):
        return 0.5 * (1.0 + erf((x - self._mu) / (self._sigma * sqrt(2.0))))

    def overlap(self, another):
        """
        Compute the overlapping coefficient (OVL) between two normal distributions.
        Measures the agreement between two normal probability distributions.
        Returns a value between 0.0 and 1.0 giving the overlapping area in
        the two underlying probability density functions.
        """ 
        # See: "The overlapping coefficient as a measure of agreement between
        # probability distributions and point estimation of the overlap of two
        # normal densities" -- Henry F. Inman and Edwin L. Bradley Jr
        # http://dx.doi.org/10.1080/03610928908830127
        X, Y = self, another
        if (Y._sigma, Y._mu) < (X._sigma, X._mu):  # sort to assure commutativity
            X, Y = Y, X
        X_var, Y_var = X.variance, Y.variance
        
        dv = Y_var - X_var
        dm = fabs(Y._mu - X._mu)
        if not dv:
            return 1.0 - erf(dm / (2.0 * X._sigma * sqrt(2.0)))
        a = X._mu * Y_var - Y._mu * X_var
        b = X._sigma * Y._sigma * sqrt(dm**2.0 + dv * log(Y_var / X_var))
        x1 = (a + b) / dv # intersection point 1
        x2 = (a - b) / dv # intersection point 2
        
        p_overlap = 1.0 - (fabs(Y.cdf(x1) - X.cdf(x1)) + fabs(Y.cdf(x2) - X.cdf(x2)))
        list_cdf_at_intersections = [X.cdf(x1),Y.cdf(x1),X.cdf(x2),Y.cdf(x2)]
        
        return p_overlap,list_cdf_at_intersections

    @property
    def mean(self):
        "Arithmetic mean of the normal distribution."
        return self._mu

    @property
    def stdev(self):
        "Standard deviation of the normal distribution."
        return self._sigma

    @property
    def variance(self):
        "Square of the standard deviation."
        return self._sigma ** 2.0




import numpy as np
import os


def makedirs(main_dir, sub_dirs=None, opt='depth'):
    if not os.path.exists(main_dir):
        os.makedirs(main_dir)
    filepaths = dict()
    if sub_dirs is not None:
        if opt == 'depth':
            path = main_dir
            for sub_dir in sub_dirs:
                path = os.path.join(path, sub_dir)
                filepaths[sub_dir] = path
                # if not os.path.exists(path):
                try:  # Use try-catch for the case of multiprocessing.
                    os.makedirs(path)
                except:
                    pass

        else:  # opt == 'breadth'
            for sub_dir in sub_dirs:
                path = os.path.join(main_dir, sub_dir)
                filepaths[sub_dir] = path
                # if not os.path.exists(path):
                try:  # Use try-catch for the case of multiprocessing.
                    os.makedirs(path)
                except:
                    pass

    return filepaths


def str_decode(df):
    str_df = df.select_dtypes([np.object])
    str_df = str_df.stack().str.decode('utf-8').unstack()
    for col in str_df:
        df[col] = str_df[col]
    return df


def str_encode(df):
    str_df = df.select_dtypes([np.object])
    str_df = str_df.stack().str.encode('utf-8').unstack()
    for col in str_df:
        df[col] = str_df[col]
    return df



import numpy as np
import pandas as pd
import os,re
import multiprocessing 
import h5py
import csv
import ujson
from operator import itemgetter
from collections import defaultdict
from io import StringIO

from . import helper
from ..utils import misc

def index(eventalign_result,pos_start,out_paths,locks):
    eventalign_result = eventalign_result.set_index(['contig','read_index'])
    pos_end=pos_start
    with locks['index'], open(out_paths['index'],'a') as f_index:
        for index in list(dict.fromkeys(eventalign_result.index)):
            transcript_id,read_index = index
            pos_end += eventalign_result.loc[index]['line_length'].sum()
            
            try: # sometimes read_index is nan
                f_index.write('%s,%d,%d,%d\n' %(transcript_id,read_index,pos_start,pos_end))
            except:
                pass
            pos_start = pos_end

def parallel_index(eventalign_filepath,chunk_size,out_dir,n_processes,resume):
    # Create output paths and locks.
    out_paths,locks = dict(),dict()
    for out_filetype in ['index']:
        out_paths[out_filetype] = os.path.join(out_dir,'eventalign.%s' %out_filetype)
        locks[out_filetype] = multiprocessing.Lock()
        
        
#     read_names_done = []
#     if resume and os.path.exists(out_paths['log']):
#         read_names_done = [line.rstrip('\n') for line in open(out_paths['log'],'r')]
#     else:
        # Create empty files.
    with open(out_paths['index'],'w') as f:
        f.write('transcript_id,read_index,pos_start,pos_end\n') # header


    # Create communication queues.
    task_queue = multiprocessing.JoinableQueue(maxsize=n_processes * 2)

    # Create and start consumers.
    consumers = [helper.Consumer(task_queue=task_queue,task_function=index,locks=locks) for i in range(n_processes)]
    for p in consumers:
        p.start()
        
    ## Load tasks into task_queue. A task is eventalign information of one read.
    eventalign_file = open(eventalign_filepath,'r')
    pos_start = len(eventalign_file.readline()) #remove header
    chunk_split = None
    index_features = ['contig','read_index','line_length']
    for chunk in pd.read_csv(eventalign_filepath, chunksize=chunk_size,sep='\t'):
        chunk_complete = chunk[chunk['read_index'] != chunk.iloc[-1]['read_index']]
        chunk_concat = pd.concat([chunk_split,chunk_complete])
        chunk_concat_size = len(chunk_concat.index)
        ## read the file at where it left off because the file is opened once ##
        lines = [len(eventalign_file.readline()) for i in range(chunk_concat_size)]
        chunk_concat['line_length'] = np.array(lines)
        task_queue.put((chunk_concat[index_features],pos_start,out_paths))
        pos_start += sum(lines)
        chunk_split = chunk[chunk['read_index'] == chunk.iloc[-1]['read_index']]
    ## the loop above leaves off w/o adding the last read_index to eventalign.index
    chunk_split_size = len(chunk_split.index)
    lines = [len(eventalign_file.readline()) for i in range(chunk_split_size)]
    chunk_split['line_length'] = np.array(lines)
    task_queue.put((chunk_split[index_features],pos_start,out_paths))

    # Put the stop task into task_queue.
    task_queue = helper.end_queue(task_queue,n_processes)

    # Wait for all of the tasks to finish.
    task_queue.join()
    
def t2g(gene_id,fasta_dict,annotation_dict,g2t_mapping,df_eventalign_index,readcount_min):
    tx_ids = []
    t2g_dict = {}
    transcripts = [tx for tx in annotation_dict if tx in g2t_mapping[gene_id]]
    n_reads = sum([len(df_eventalign_index.loc[tx]) for tx in transcripts])
    if n_reads >= readcount_min:
        for tx in transcripts:
            tx_seq = fasta_dict[tx][0]
            tx_contig = annotation_dict[tx]['chr']
            if tx_seq is None:
                continue
            for exon_num in range(len(annotation_dict[tx]['exon'])):
                g_interval=annotation_dict[tx]['exon'][exon_num]
                tx_interval=annotation_dict[tx]['tx_exon'][exon_num]
                for g_pos in range(g_interval[0],g_interval[1]+1): # Exclude the rims of exons.
                    dis_from_start = g_pos - g_interval[0]
                    if annotation_dict[tx]['strand'] == "+":
                        tx_pos = tx_interval[0] + dis_from_start
                    elif annotation_dict[tx]['strand'] == "-":
                        tx_pos = tx_interval[1] - dis_from_start
                    if (g_interval[0] <= g_pos < g_interval[0]+2) or (g_interval[1]-2 < g_pos <= g_interval[1]): # Todo: To improve the mapping
                        kmer = 'XXXXX'
                    else:
                        kmer = tx_seq[tx_pos-2:tx_pos+3]
                    t2g_dict[(tx,tx_pos)] = (tx_contig,gene_id,g_pos,kmer) # tx.contig is chromosome.
            tx_ids += [tx]

                
    return n_reads, tx_ids, t2g_dict

def combine(events_str):
    f_string = StringIO(events_str)
    eventalign_result = pd.read_csv(f_string,delimiter='\t',names=['contig','position','reference_kmer','read_index',
                         'strand','event_index','event_level_mean','event_stdv','event_length','model_kmer',
                         'model_mean', 'model_stdv', 'standardized_level', 'start_idx', 'end_idx'])
    f_string.close()
    cond_successfully_eventaligned = eventalign_result['reference_kmer'] == eventalign_result['model_kmer']
    if cond_successfully_eventaligned.sum() != 0:

        eventalign_result = eventalign_result[cond_successfully_eventaligned]

        keys = ['read_index','contig','position','reference_kmer'] # for groupby
        eventalign_result['length'] = pd.to_numeric(eventalign_result['end_idx'])-pd.to_numeric(eventalign_result['start_idx'])
        eventalign_result['sum_norm_mean'] = pd.to_numeric(eventalign_result['event_level_mean']) * eventalign_result['length']
            
        eventalign_result = eventalign_result.groupby(keys)  
        sum_norm_mean = eventalign_result['sum_norm_mean'].sum() 
        start_idx = eventalign_result['start_idx'].min()
        end_idx = eventalign_result['end_idx'].max()
        total_length = eventalign_result['length'].sum()

        eventalign_result = pd.concat([start_idx,end_idx],axis=1)
        eventalign_result['norm_mean'] = (sum_norm_mean/total_length).round(1)

        eventalign_result.reset_index(inplace=True)


#        eventalign_result['transcript_id'] = [contig.split('.')[0] for contig in eventalign_result['contig']]    #### CHANGE MADE ####
        eventalign_result['transcript_id'] = [contig for contig in eventalign_result['contig']]
        #eventalign_result['transcript_id'] = eventalign_result['contig']

        eventalign_result['transcriptomic_position'] = pd.to_numeric(eventalign_result['position']) + 2 # the middle position of 5-mers.
        # eventalign_result = misc.str_encode(eventalign_result)
#         eventalign_result['read_id'] = [read_name]*len(eventalign_result)

        # features = ['read_id','transcript_id','transcriptomic_position','reference_kmer','norm_mean','start_idx','end_idx']
        # features_dtype = np.dtype([('read_id', 'S36'), ('transcript_id', 'S15'), ('transcriptomic_position', '<i8'), ('reference_kmer', 'S5'), ('norm_mean', '<f8'), ('start_idx', '<i8'), ('end_idx', '<i8')])
        
#         features = ['transcript_id','transcriptomic_position','reference_kmer','norm_mean']

#         df_events = eventalign_result[['read_index']+features]
#         # print(df_events.head())

        features = ['transcript_id','transcriptomic_position','reference_kmer','norm_mean']
#        np_events = eventalign_result[features].reset_index().values.ravel().view(dtype=[('transcript_id', 'S15'), ('transcriptomic_position', '<i8'), ('reference_kmer', 'S5'), ('norm_mean', '<f8')])
        df_events = eventalign_result[features]
        np_events = np.rec.fromrecords(df_events, names=[*df_events])
        return np_events

def readFasta(transcript_fasta,is_gff):
    fasta=open(transcript_fasta,"r")
    entries,separate_by_pipe="",False
    for ln in fasta:
        entries+=ln
    entries=entries.split(">")
    if len(entries[1].split("|"))>1:
        separate_by_pipe=True
    dict={}
    for entry in entries:
        entry=entry.split("\n")
        if len(entry[0].split()) > 0:
            id=entry[0].split('.')[0]
            seq="".join(entry[1:])
            dict[id]=[seq]
            if is_gff > 0:
                if separate_by_pipe == True:
                    g_id=info[1],split(".")[0]
                else:
                    g_id=entry[0].split("gene:")[1].split(".")[0]
                dict[id].append(g_id)
    return dict

def readAnnotation(gtf_or_gff):
    gtf=open(gtf_or_gff,"r")
    dict,is_gff={},0
    for ln in gtf:
        if not ln.startswith("#"):
            ln=ln.strip("\n").split("\t")
            if is_gff == 0:
                if ln[-1].startswith("ID") or ln[-1].startswith("Parent"):
                    is_gff = 1
                else:
                    is_gff = -1
            if is_gff < 0:
                if ln[2] == "transcript" or ln[2] == "exon":
                    chr,type,start,end=ln[0],ln[2],int(ln[3]),int(ln[4])
                    attrList=ln[-1].split(";")
                    attrDict={}
                    for k in attrList:
                        p=k.strip().split(" ")
                        if len(p) == 2:
                            attrDict[p[0]]=p[1].strip('\"')
                    ##tx_id=ln[-1].split('; transcript_id "')[1].split('";')[0]
                    ##g_id=ln[-1].split('gene_id "')[1].split('";')[0]
                    tx_id = attrDict["transcript_id"]
                    g_id = attrDict["gene_id"]
                    if tx_id not in dict:
                        dict[tx_id]={'chr':chr,'g_id':g_id,'strand':ln[6]}
                        if type not in dict[tx_id]:
                            if type == "transcript":
                                dict[tx_id][type]=(start,end)
                    else:
                        if type == "exon":
                            if type not in dict[tx_id]:
                                dict[tx_id][type]=[(start,end)]
                            else:
                                dict[tx_id][type].append((start,end))
            if is_gff > 0:
                if ln[2] == "exon" or ln[2] == "mRNA":
                    chr,type,start,end=ln[0],ln[2],int(ln[3]),int(ln[4])
                    tx_id=ln[-1].split('transcript:')[1].split(';')[0]
                    if ln[2] == "mRNA":
                        type="transcript"
                    if tx_id not in dict:
                        dict[tx_id]={'chr':chr,'strand':ln[6]}
                        if type == "transcript":
                            dict[tx_id][type]=(start,end)
                        if type == "exon":
                            dict[tx_id][type]=[(start,end)]
                    else:
                        if type == "transcript" and type not in dict[tx_id]:
                            dict[tx_id][type]=(start,end)
                        if type == "exon":
                            if type not in dict[tx_id]:
                                dict[tx_id][type]=[(start,end)]
                            else:
                                dict[tx_id][type].append((start,end))
    #convert genomic positions to tx positions
    if is_gff < 0:
        for id in dict:
            tx_pos,tx_start=[],0
            for pair in dict[id]["exon"]:
                tx_end=pair[1]-pair[0]+tx_start
                tx_pos.append((tx_start,tx_end))
                tx_start=tx_end+1
            dict[id]['tx_exon']=tx_pos
    else:
        for id in dict:
            tx_pos,tx_start=[],0
            if dict[id]["strand"] == "-":
                dict[id]["exon"].sort(key=lambda tup: tup[0], reverse=True)
            for pair in dict[id]["exon"]:
                tx_end=pair[1]-pair[0]+tx_start
                tx_pos.append((tx_start,tx_end))
                tx_start=tx_end+1
            dict[id]['tx_exon']=tx_pos
    return (dict,is_gff)

def parallel_preprocess_gene(eventalign_filepath,fasta_dict,annotation_dict,is_gff,out_dir,n_processes,readcount_min,readcount_max,resume):
    
    # Create output paths and locks.
    out_paths,locks = dict(),dict()
    for out_filetype in ['json','index','log','readcount']:
        out_paths[out_filetype] = os.path.join(out_dir,'data.%s' %out_filetype)
        locks[out_filetype] = multiprocessing.Lock()
                
    # Writing the starting of the files.
    gene_ids_done = []
    if resume and os.path.exists(out_paths['index']):
        df_index = pd.read_csv(out_paths['index'],sep=',')
        gene_ids_done = list(df_index['idx'].unique())
    else:
        # with open(out_paths['json'],'w') as f:
        #     f.write('{\n')
        #     f.write('"genes":{')
        open(out_paths['json'],'w').close()
        with open(out_paths['index'],'w') as f:
            f.write('idx,start,end\n') # header
        with open(out_paths['readcount'],'w') as f:
            f.write('idx,n_reads\n') # header
        open(out_paths['log'],'w').close()

    # Create communication queues.
    task_queue = multiprocessing.JoinableQueue(maxsize=n_processes * 2)

    # Create and start consumers.
    consumers = [helper.Consumer(task_queue=task_queue,task_function=preprocess_gene,locks=locks) for i in range(n_processes)]
    for p in consumers:
        p.start()

    # Get all gene ids and create a dict of eventalign.combine index.
#     gene_ids = set()

    if is_gff > 0: ##add g_id from fasta dict entry if gff annotation is used
        for tx_id in annotation_dict:
            try:
                annotation_dict[tx_id]['g_id']=fasta_dict[tx_id][1]
            except KeyError:
                continue

    df_eventalign_index = pd.read_csv(os.path.join(out_dir,'eventalign.index'))
    df_eventalign_index['transcript_id'] = [tx_id.split('.')[0] for tx_id in  df_eventalign_index['transcript_id']]
#    df_eventalign_index['transcript_id'] = [tx_id for tx_id in  df_eventalign_index['transcript_id']]
    df_eventalign_index.set_index('transcript_id',inplace=True)
    g2t_mapping = defaultdict(list)

    for tx_id in set(df_eventalign_index.index):
        try:
##           g_id = ensembl.transcript_by_id(tx_id).gene_id 
            g_id = annotation_dict[tx_id]['g_id'] 
        except KeyError:
            continue
        else:
#             gene_ids = gene_ids.union([g_id])
            g2t_mapping[g_id] += [tx_id]

#     f = open(os.path.join(out_dir,'eventalign.index'))
#     for ln in f:
#         tx_id,read_index,pos_start,pos_end = ln.split(',')
#         tx_id,tx_version = tx_id.split('.') # Based on Ensembl
#         eventalign_index[tx_id] += [(int(read_index),int(pos_start),int(pos_end))]
#         tx_ensembl[tx_id] = tx_version
#         try:
#             g_id = ensembl.transcript_by_id(tx_id).gene_id 
#         except ValueError:
#             continue
#         else:
#             gene_ids = gene_ids.union([g_id])
            

    # Load tasks into task_queue.    
    gene_ids_processed = []

    with open(eventalign_filepath,'r') as eventalign_result:

        for gene_id in g2t_mapping:
                        
            if resume and (gene_id in gene_ids_done):
                continue
            # mapping a gene <-> transcripts

            n_reads, tx_ids, t2g_mapping = t2g(gene_id,fasta_dict,annotation_dict,g2t_mapping,df_eventalign_index,readcount_min)
            #
            if n_reads >= readcount_min: 
                data_dict = dict()
                readcount = 0
                for tx_id in tx_ids:
                    for _,row in df_eventalign_index.loc[[tx_id]].iterrows():
                        read_index,pos_start,pos_end = row['read_index'],row['pos_start'],row['pos_end']
                        eventalign_result.seek(pos_start,0)
                        events_str = eventalign_result.read(pos_end-pos_start)
                        data = combine(events_str)
                        #data = np.genfromtxt(f_string,delimiter=',',dtype=np.dtype([('transcript_id', 'S15'), ('transcriptomic_position', '<i8'), ('reference_kmer', 'S5'), ('norm_mean', '<f8')]))
                        if (data is not None) and (data.size > 1):
                            data_dict[read_index] = data
                        readcount += 1 
                        if readcount > readcount_max:
                            break

                    if readcount > readcount_max:
                        break
                if len(data_dict)>=readcount_min:
#                     print(gene_id,len(data_dict)) #len(data_dict) is the number of reads to be processed.
                    task_queue.put((gene_id,data_dict,t2g_mapping,out_paths)) # Blocked if necessary until a free slot is available. 
                    gene_ids_processed += [gene_id]


    # Put the stop task into task_queue.
    task_queue = helper.end_queue(task_queue,n_processes)

    # Wait for all of the tasks to finish.
    task_queue.join()
    
    with open(out_paths['log'],'a+') as f:
        f.write('Total %d genes.\n' %len(gene_ids_processed))
        f.write(helper.decor_message('successfully finished'))

def preprocess_gene(gene_id,data_dict,t2g_mapping,out_paths,locks):  
    """
    Convert transcriptomic to genomic coordinates for a gene.
    
    Parameters
    ----------
        gene_id: str
            Gene ID.
        data_dict: {tx_id:events_array}
            Events for each read.
        t2g_mapping: {(,):()}
            A dict to map transcriptomic coordinates (transcript id and transcriptomic position) to genomic (gene id and genomic position).
        db_type: 
            Type of gene-tx mapping either EnsemblRelease or (customised) Genome 
        features: [str] # todo
            A list of features to collect from the reads that are aligned to each genomic coordinate in the output.
    Returns
    -------
    dict
        A dict of all specified features collected for each genomic coordinate.
    """
    
    # features = ['read_id','transcript_id','transcriptomic_position','reference_kmer','norm_mean','start_idx','end_idx'] # columns in the eventalign file per read.

    events = []
    condition_labels = []
    run_labels = []
    read_ids = []
    genomic_coordinates = []
    
    # Concatenate
#     if len(data_dict) == 0:
#         return


    for read_index,events_per_read in data_dict.items():
#         if len(events_per_read) > 0:
        # ===== transcript to gene coordinates ===== # TODO: to use gtf.
#        tx_ids = [tx_id.decode('UTF-8').split('.')[0] for tx_id in events_per_read['transcript_id']]
        tx_ids = [tx_id.split('.')[0] for tx_id in events_per_read['transcript_id']] 
        tx_positions = events_per_read['transcriptomic_position']
        genomic_coordinate = list(itemgetter(*zip(tx_ids,tx_positions))(t2g_mapping)) # genomic_coordinates -- np structured array of 'chr','gene_id','genomic_position','kmer'
        genomic_coordinate = np.array(genomic_coordinate,dtype=np.dtype([('chr','<U2'),('gene_id','<U15'),('genomic_position','<i4'),('g_kmer','<U5')]))
        # ===== 

        # Based on Ensembl, remove transcript version.

        events_per_read['transcript_id'] = tx_ids
        events_per_read = np.array(events_per_read,dtype=np.dtype([('transcript_id', 'S15'), ('transcriptomic_position', '<i8'), ('reference_kmer', 'S5'), ('norm_mean', '<f8')]))

        #

        events += [events_per_read]
        genomic_coordinates += [genomic_coordinate]
        n_events_per_read = len(events_per_read)
#         else:
#             print(read_index,len(events_per_read))

    events = np.concatenate(events)
    genomic_coordinates = np.concatenate(genomic_coordinates)
   
    # Sort and split # 
#     idx_sorted = np.lexsort((events['reference_kmer'],genomic_coordinates['genomic_position'],genomic_coordinates['gene_id']))
#     key_tuples, index = np.unique(list(zip(genomic_coordinates['gene_id'][idx_sorted],genomic_coordinates['genomic_position'][idx_sorted],events['reference_kmer'][idx_sorted])),return_index = True,axis=0) #'chr',
#     y_arrays = np.split(events['norm_mean'][idx_sorted], index[1:])
# #     read_id_arrays = np.split(events['read_id'][idx_sorted], index[1:])
#     g_kmer_arrays = np.split(genomic_coordinates['g_kmer'][idx_sorted], index[1:])

    idx_sorted = np.argsort(genomic_coordinates['genomic_position'])
    unique_positions, index = np.unique(genomic_coordinates['genomic_position'][idx_sorted],return_index = True)
    y_arrays = np.split(events['norm_mean'][idx_sorted], index[1:])
    #     read_id_arrays = np.split(events['read_id'][idx_sorted], index[1:])
    g_kmer_arrays = np.split(genomic_coordinates['g_kmer'][idx_sorted], index[1:])
    g_positions_arrays = np.split(genomic_coordinates['genomic_position'][idx_sorted], index[1:])

    # Prepare
    # print('Reformating the data for each genomic position ...')
    data = defaultdict(dict)
    # for each position, make it ready for json dump
#     data = dict(zip(key_tuples, y_arrays))

    asserted = True
#     for key_tuple,y_array,g_kmer_array in zip(key_tuples,y_arrays,g_kmer_arrays):
    for position,y_array,g_kmer_array,g_positions_array in zip(unique_positions,y_arrays,g_kmer_arrays,g_positions_arrays):
#         gene_id,position,kmer = key_tuple            
        if (len(set(g_kmer_array)) == 1) and ('XXXXX' in set(g_kmer_array)) or (len(y_array) == 0):
            continue
            
        if 'XXXXX' in set(g_kmer_array):
            y_array = y_array[g_kmer_array != 'XXXXX']  
            assert len(y_array) == len(g_kmer_array) - (g_kmer_array=='XXXXX').sum()
            g_kmer_array = g_kmer_array[g_kmer_array != 'XXXXX']  
            
        try:
            assert len(set(g_kmer_array)) == 1
            assert list(set(g_kmer_array))[0].count('N') == 0 ##to weed out the mapped kmers from tx_seq that contain 'N', which is not in diffmod's model_kmer
            assert {position} == set(g_positions_array)
        except:
            asserted = False
            break
        kmer = set(g_kmer_array).pop()

        data[position] = {kmer: list(y_array)} #,'read_ids': [read_id.decode('UTF-8') for read_id in read_id_array]}
        
    # write to file.
    log_str = '%s: %s' %(gene_id,asserted)

    with locks['json'], open(out_paths['json'],'a') as f:

        pos_start = f.tell()
        f.write('{')
        f.write('"%s":' %gene_id)
        ujson.dump(data, f)
        f.write('}\n')
        pos_end = f.tell()

    with locks['index'], open(out_paths['index'],'a') as f:
        f.write('%s,%d,%d\n' %(gene_id,pos_start,pos_end))
        
    with locks['readcount'], open(out_paths['readcount'],'a') as f: #todo: repeats no. of tx >> don't want it.
        n_reads = len(data_dict)
        f.write('%s,%d\n' %(gene_id,n_reads))
        
    with locks['log'], open(out_paths['log'],'a') as f:
        f.write(log_str + '\n')


def parallel_preprocess_tx(eventalign_filepath,out_dir,n_processes,readcount_min,readcount_max,resume):
    
    # Create output paths and locks.
    out_paths,locks = dict(),dict()
    for out_filetype in ['json','index','log','readcount']:
        out_paths[out_filetype] = os.path.join(out_dir,'data.%s' %out_filetype)
        locks[out_filetype] = multiprocessing.Lock()
                
    # Writing the starting of the files.
    tx_ids_done = []
    if resume and os.path.exists(out_paths['index']):
        df_index = pd.read_csv(out_paths['index'],sep=',')
        tx_ids_done = list(df_index['transcript_id'].unique())
    else:
        open(out_paths['json'],'w').close()
        with open(out_paths['index'],'w') as f:
            f.write('idx,start,end\n') # header
        with open(out_paths['readcount'],'w') as f:
            f.write('idx,n_reads\n') # header
        open(out_paths['log'],'w').close()

    # Create communication queues.
    task_queue = multiprocessing.JoinableQueue(maxsize=n_processes * 2)

    # Create and start consumers.
    consumers = [helper.Consumer(task_queue=task_queue,task_function=preprocess_tx,locks=locks) for i in range(n_processes)]
    for p in consumers:
        p.start()

    # Load tasks into task_queue.
    tx_ids_processed = []
    df_eventalign_index = pd.read_csv(os.path.join(out_dir,'eventalign.index'))
#    df_eventalign_index['transcript_id'] = [tx_id.split('.')[0] for tx_id in  df_eventalign_index['transcript_id']]
#    df_eventalign_index['transcript_id'] = [tx_id for tx_id in  df_eventalign_index['transcript_id']]
    tx_ids = df_eventalign_index['transcript_id'].values.tolist()
    tx_ids = list(dict.fromkeys(tx_ids))
    df_eventalign_index.set_index('transcript_id',inplace=True)
    with open(eventalign_filepath,'r') as eventalign_result:
        for tx_id in tx_ids:
            data_dict = dict()
            readcount = 0
            for _,row in df_eventalign_index.loc[[tx_id]].iterrows():
                read_index,pos_start,pos_end = row['read_index'],row['pos_start'],row['pos_end']
                eventalign_result.seek(pos_start,0)
                events_str = eventalign_result.read(pos_end-pos_start)
                data = combine(events_str)
                if (data is not None) and (data.size > 1):
                    data_dict[read_index] = data
                readcount += 1 
                if readcount > readcount_max:
                    break
            if readcount>=readcount_min:
                task_queue.put((tx_id,data_dict,out_paths)) # Blocked if necessary until a free slot is available. 
                tx_ids_processed += [tx_id]

    # Put the stop task into task_queue.
    task_queue = helper.end_queue(task_queue,n_processes)

    # Wait for all of the tasks to finish.
    task_queue.join()
    
    with open(out_paths['log'],'a+') as f:
        f.write('Total %d transcripts.\n' %len(tx_ids_processed))
        f.write(helper.decor_message('successfully finished'))

def preprocess_tx(tx_id,data_dict,out_paths,locks): 
    """
    Convert transcriptomic to genomic coordinates for a gene.
    
    Parameters
    ----------
        tx_id: str
            Transcript ID.
        data_dict: {read_id:events_array}
            Events for each read.
        features: [str] # todo
            A list of features to collect from the reads that are aligned to each genomic coordinate in the output.
    Returns
    -------
    dict
        A dict of all specified features collected for each genomic coordinate.
    """
    
    # features = ['read_id','transcript_id','transcriptomic_position','reference_kmer','norm_mean','start_idx','end_idx'] # columns in the eventalign file per read.

    events = []
    condition_labels = []
    run_labels = []
    read_ids = []
    transcriptomic_coordinates = []
    
    # Concatenate
    if len(data_dict) == 0:
        return

    for read_id,events_per_read in data_dict.items(): 
        # print(read_id)
        events += [events_per_read]
        
    events = np.concatenate(events)
   
    # Sort and split 
    idx_sorted = np.argsort(events['transcriptomic_position'])
    unique_positions, index = np.unique(events['transcriptomic_position'][idx_sorted],return_index = True)
    y_arrays = np.split(events['norm_mean'][idx_sorted], index[1:])
#    read_id_arrays = np.split(events['read_id'][idx_sorted], index[1:])
    reference_kmer_arrays = np.split(events['reference_kmer'][idx_sorted], index[1:])

    # Prepare
    # print('Reformating the data for each genomic position ...')
    data = defaultdict(dict)
    # for each position, make it ready for json dump
    asserted = True
#     for key_tuple,y_array,reference_kmer_array in zip(key_tuples,y_arrays,reference_kmer_arrays):
    for position,y_array,reference_kmer_array in zip(unique_positions,y_arrays,reference_kmer_arrays):
        
        position = int(position)
        if (len(set(reference_kmer_array)) == 1) and ('XXXXX' in set(reference_kmer_array)) or (len(y_array) == 0):
            continue
            
        if 'XXXXX' in set(reference_kmer_array):
            y_array = y_array[reference_kmer_array != 'XXXXX']  
            assert len(y_array) == len(reference_kmer_array) - (reference_kmer_array=='XXXXX').sum()
            reference_kmer_array = reference_kmer_array[reference_kmer_array != 'XXXXX']  
            
        try:
            assert len(set(reference_kmer_array)) == 1
            assert list(set(reference_kmer_array))[0].count('N') == 0 ##to weed out the mapped kmers from tx_seq that contain 'N', which is not in diffmod's model_kmer
        except:
            asserted = False
            break
        kmer = set(reference_kmer_array).pop()

        data[position] = {kmer: list(np.around(y_array,decimals=2))}
        
    # write to file.
    log_str = '%s: %s.' %(tx_id,asserted)
    with locks['json'], open(out_paths['json'],'a') as f:
        pos_start = f.tell()
        f.write('{')
        f.write('"%s":' %tx_id)
        ujson.dump(data, f)
        f.write('}\n')
        pos_end = f.tell()
        
    with locks['index'], open(out_paths['index'],'a') as f:
        f.write('%s,%d,%d\n' %(tx_id,pos_start,pos_end))
        
    with locks['readcount'], open(out_paths['readcount'],'a') as f: #todo: repeats no. of tx >> don't want it.
        n_reads = len(data_dict)
        f.write('%s,%d\n' %(tx_id,n_reads))
        
    with locks['log'], open(out_paths['log'],'a') as f:
        f.write(log_str + '\n')
        

# def index_nanopolish(eventalign_filepath,summary_filepath,out_dir,n_processes):
#     with helper.EventalignFile(eventalign_filepath) as eventalign_file, open(summary_filepath,'r') as summary_file:
        
#         reader_summary = csv.DictReader(summary_file, delimiter="\t")
#         reader_eventalign = csv.DictReader(eventalign_file, delimiter="\t")

#         row_summary = next(reader_summary)
#         read_name = row_summary['read_name']
#         read_index = row_summary['read_index']
#         eventalign_per_read = []
#         for row_eventalign in reader_eventalign:
#             if (row_eventalign['read_index'] == read_index):
#                 eventalign_per_read += [row_eventalign]
#             else: 
#                 # Load a read info to the task queue.
#                 if read_name not in read_names_done:
#                     task_queue.put((read_name,eventalign_per_read,out_paths))
#                 # Next read.
#                 try:
#                     row_summary = next(reader_summary)
#                 except StopIteration: # no more read.
#                     break
#                 else:
#                     read_index = row_summary['read_index']
#                     read_name = row_summary['read_name']
#                     assert row_eventalign['read_index'] == read_index 
#                     eventalign_per_read = [row_eventalign]

#def check_gene_tx_id_version(gtf_or_gff):
#    gtf=open(gtf_or_gff,"r")
#    extra_version_fields=0
#    for i in range(25):
#        ln=gtf.readline().split('\t')
#        if not ln[0].startswith('#'):
#            if ln[2] == "transcript" or ln[2] == "exon":
#                check_transcript_version = len(ln[-1].split('transcript_version')) == 2
#                check_gene_version = len(ln[-1].split('gene_version')) == 2
#                if check_transcript_version and check_gene_version:
#                   extra_version_fields+=1
#    if extra_version_fields > 0:
#        return True
#    else:
#        return False
 
#def mergeGTFtxIDversion(gtf_or_gff,out_dir):
#    gtf=open(gtf_or_gff,"r")
#    new_gtf_path=os.path.join(out_dir,'transcript_id_version_merged.gtf')
#    new_gtf=open(new_gtf_path,"w")
#    for ln in gtf:
#        if not ln.startswith("#"):
#            ln=ln.split("\t")
#            if ln[2] == "transcript" or ln[2] == "exon":
#                g_id=ln[-1].split('gene_id "')[1].split('";')[0]
#                g_ver=ln[-1].split('; gene_version "')[1].split('";')[0]
#                tx_id=ln[-1].split('; transcript_id "')[1].split('";')[0]
#                tx_ver=ln[-1].split('; transcript_version "')[1].split('";')[0]
#                new_gtf.write('\t'.join(ln[:-1])+'\t'+'gene_id "'+g_id+'.'+g_ver+'"; transcript_id "'+tx_id+'.'+tx_ver+'";'+'\n')
#    new_gtf.close()
#    return new_gtf_path

def dataprep(args):
    #
    n_processes = args.n_processes        
    eventalign_filepath = args.eventalign
    chunk_size = args.chunk_size
    out_dir = args.out_dir
    readcount_min = args.readcount_min    
    readcount_max = args.readcount_max
    resume = args.resume
    genome = args.genome

    if genome and (None in [args.gtf_or_gff,args.transcript_fasta]):
        print('please provide the following')
        print('- gtf_or_gff')
        print('- transcript_fasta')
    else:
        gtf_or_gff = args.gtf_or_gff
        transcript_fasta = args.transcript_fasta
        
    misc.makedirs(out_dir) #todo: check every level.
    
    # (1) For each read, combine multiple events aligned to the same positions, the results from nanopolish eventalign, into a single event per position.
    if not args.skip_eventalign_indexing:
        parallel_index(eventalign_filepath,chunk_size,out_dir,n_processes,resume)
    
    # (2) Create a .json file, where the info of all reads are stored per position, for modelling.
    if genome:
#        merge_transcript_id_version = check_gene_tx_id_version(gtf_or_gff)
#        if merge_transcript_id_version:
#            gtf_or_gff = mergeGTFtxIDversion(gtf_or_gff,out_dir)
        annotation_dict,is_gff = readAnnotation(gtf_or_gff)
        fasta_dict = readFasta(transcript_fasta,is_gff)
        parallel_preprocess_gene(eventalign_filepath,fasta_dict,annotation_dict,is_gff,out_dir,n_processes,readcount_min,readcount_max,resume)
    else:
        parallel_preprocess_tx(eventalign_filepath,out_dir,n_processes,readcount_min,readcount_max,resume)



"""Setup for the xpore package."""

from setuptools import setup,find_packages

__pkg_name__ = 'xpore'


with open('README.md') as f:
    README = f.read()

setup(
    author="Ploy N. Pratanwanich",
    maintainer_email="naruemon.p@chula.ac.th",
    name=__pkg_name__,
    license="MIT",
    description='xpore is a python package for Nanopore data analysis of differential RNA modifications.',
    version='v2.1',
    long_description=README,
    long_description_content_type='text/markdown',
    url='https://github.com/GoekeLab/xpore',
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
            'numpy>=1.18.0',
            'pandas>=0.25.3',
            'scipy>=1.4.1',
            'PyYAML',
            'h5py>=2.10.0',
            'pyensembl>=1.8.5',
            'ujson>=4.0.1'
            ],
    python_requires=">=3.8",
    entry_points={'console_scripts': ["xpore={}.scripts.xpore:main".format(__pkg_name__)]},
    classifiers=[
        # Trove classifiers
        # (https://pypi.python.org/pypi?%3Aaction=list_classifiers)
        'Development Status :: 1 - Planning',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.8',
        'Topic :: Software Development :: Libraries',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Intended Audience :: Science/Research',
    ],
)



import scipy.stats
import numpy as np

class StatsTest(object):
    def __init__(self,data):
        if self.__isok(data):
            self.data = [data['y'][data['x'][:,0]==1],data['y'][data['x'][:,1]==1]] 
        else:
            self.data = None

    def fit(self,method):
        pval = np.nan
        if self.data is not None:
            if method == 't-test':
                stats,_ = scipy.stats.ttest_ind(self.data[0],self.data[1])
                df = len(self.data[0]) + len(self.data[1]) - 2
                pval = scipy.stats.t.sf(np.abs(stats), df)*2 # also defined as 1 - cdf, but sf is sometimes more accurate. #(1 - scipy.stats.t.cdf(abs(stat), df)) * 
        return pval
    
    def __isok(self,data):
        return data['x'].shape[1] == 2 # Check if only two conditions.





import os

def run_postprocessing(diffmod_table_path,out_dir):
    file=open(diffmod_table_path,"r")
    header=file.readline()
    entries=file.readlines()
    outfile_path=os.path.join(out_dir,"majority_direction_kmer_diffmod.table")
    outfile=open(outfile_path,"w")
    outfile.write(header)
    header=header.strip().split(',')
    kmer_ind,dir_ind=header.index('kmer'),header.index('mod_assignment')    
    dict={}
    for ln in entries:
        l=ln.strip().split(",")
        if l[kmer_ind] not in dict:
            dict[l[kmer_ind]]={l[dir_ind]:1}
        else:
            if l[dir_ind] not in dict[l[kmer_ind]]:
                dict[l[kmer_ind]][l[dir_ind]]=1
            else:
                dict[l[kmer_ind]][l[dir_ind]]+=1
    for k in dict:
        if len(dict[k]) > 1:  ##consider one modification type per k-mer
            if dict[k]['higher'] <= dict[k]['lower']: ##choose the majority
                dict[k]['choose']='lower'
            else:
                dict[k]['choose']='higher'
        else:
            dict[k]['choose']=list(dict[k].keys())[0]
    for ln in entries:
        l=ln.strip().split(",")
        if l[dir_ind] == dict[l[kmer_ind]]['choose']:
            outfile.write(ln)
    outfile.close()

def postprocessing(args):
    diffmod_dir = args.diffmod_dir
    diffmod_table_path = os.path.join(diffmod_dir,"diffmod.table")
    run_postprocessing(diffmod_table_path,diffmod_dir)

